# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: data_points.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='data_points.proto',
  package='com.cognite.v1.timeseries.proto',
  syntax='proto3',
  serialized_options=b'P\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11\x64\x61ta_points.proto\x12\x1f\x63om.cognite.v1.timeseries.proto\"4\n\x10NumericDatapoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x01\"Z\n\x11NumericDatapoints\x12\x45\n\ndatapoints\x18\x01 \x03(\x0b\x32\x31.com.cognite.v1.timeseries.proto.NumericDatapoint\"3\n\x0fStringDatapoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\t\"X\n\x10StringDatapoints\x12\x44\n\ndatapoints\x18\x01 \x03(\x0b\x32\x30.com.cognite.v1.timeseries.proto.StringDatapoint\"\xee\x01\n\x12\x41ggregateDatapoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x0f\n\x07\x61verage\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x12\x0b\n\x03min\x18\x04 \x01(\x01\x12\r\n\x05\x63ount\x18\x05 \x01(\x01\x12\x0b\n\x03sum\x18\x06 \x01(\x01\x12\x15\n\rinterpolation\x18\x07 \x01(\x01\x12\x19\n\x11stepInterpolation\x18\x08 \x01(\x01\x12\x1a\n\x12\x63ontinuousVariance\x18\t \x01(\x01\x12\x18\n\x10\x64iscreteVariance\x18\n \x01(\x01\x12\x16\n\x0etotalVariation\x18\x0b \x01(\x01\"^\n\x13\x41ggregateDatapoints\x12G\n\ndatapoints\x18\x01 \x03(\x0b\x32\x33.com.cognite.v1.timeseries.proto.AggregateDatapointB\x02P\x01\x62\x06proto3'
)




_NUMERICDATAPOINT = _descriptor.Descriptor(
  name='NumericDatapoint',
  full_name='com.cognite.v1.timeseries.proto.NumericDatapoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='com.cognite.v1.timeseries.proto.NumericDatapoint.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.cognite.v1.timeseries.proto.NumericDatapoint.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=54,
  serialized_end=106,
)


_NUMERICDATAPOINTS = _descriptor.Descriptor(
  name='NumericDatapoints',
  full_name='com.cognite.v1.timeseries.proto.NumericDatapoints',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='datapoints', full_name='com.cognite.v1.timeseries.proto.NumericDatapoints.datapoints', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=108,
  serialized_end=198,
)


_STRINGDATAPOINT = _descriptor.Descriptor(
  name='StringDatapoint',
  full_name='com.cognite.v1.timeseries.proto.StringDatapoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='com.cognite.v1.timeseries.proto.StringDatapoint.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.cognite.v1.timeseries.proto.StringDatapoint.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=200,
  serialized_end=251,
)


_STRINGDATAPOINTS = _descriptor.Descriptor(
  name='StringDatapoints',
  full_name='com.cognite.v1.timeseries.proto.StringDatapoints',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='datapoints', full_name='com.cognite.v1.timeseries.proto.StringDatapoints.datapoints', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=253,
  serialized_end=341,
)


_AGGREGATEDATAPOINT = _descriptor.Descriptor(
  name='AggregateDatapoint',
  full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='average', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.average', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.max', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.min', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.count', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sum', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.sum', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='interpolation', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.interpolation', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stepInterpolation', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.stepInterpolation', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='continuousVariance', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.continuousVariance', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='discreteVariance', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.discreteVariance', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='totalVariation', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoint.totalVariation', index=10,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=344,
  serialized_end=582,
)


_AGGREGATEDATAPOINTS = _descriptor.Descriptor(
  name='AggregateDatapoints',
  full_name='com.cognite.v1.timeseries.proto.AggregateDatapoints',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='datapoints', full_name='com.cognite.v1.timeseries.proto.AggregateDatapoints.datapoints', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=584,
  serialized_end=678,
)

_NUMERICDATAPOINTS.fields_by_name['datapoints'].message_type = _NUMERICDATAPOINT
_STRINGDATAPOINTS.fields_by_name['datapoints'].message_type = _STRINGDATAPOINT
_AGGREGATEDATAPOINTS.fields_by_name['datapoints'].message_type = _AGGREGATEDATAPOINT
DESCRIPTOR.message_types_by_name['NumericDatapoint'] = _NUMERICDATAPOINT
DESCRIPTOR.message_types_by_name['NumericDatapoints'] = _NUMERICDATAPOINTS
DESCRIPTOR.message_types_by_name['StringDatapoint'] = _STRINGDATAPOINT
DESCRIPTOR.message_types_by_name['StringDatapoints'] = _STRINGDATAPOINTS
DESCRIPTOR.message_types_by_name['AggregateDatapoint'] = _AGGREGATEDATAPOINT
DESCRIPTOR.message_types_by_name['AggregateDatapoints'] = _AGGREGATEDATAPOINTS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NumericDatapoint = _reflection.GeneratedProtocolMessageType('NumericDatapoint', (_message.Message,), {
  'DESCRIPTOR' : _NUMERICDATAPOINT,
  '__module__' : 'data_points_pb2'
  # @@protoc_insertion_point(class_scope:com.cognite.v1.timeseries.proto.NumericDatapoint)
  })
_sym_db.RegisterMessage(NumericDatapoint)

NumericDatapoints = _reflection.GeneratedProtocolMessageType('NumericDatapoints', (_message.Message,), {
  'DESCRIPTOR' : _NUMERICDATAPOINTS,
  '__module__' : 'data_points_pb2'
  # @@protoc_insertion_point(class_scope:com.cognite.v1.timeseries.proto.NumericDatapoints)
  })
_sym_db.RegisterMessage(NumericDatapoints)

StringDatapoint = _reflection.GeneratedProtocolMessageType('StringDatapoint', (_message.Message,), {
  'DESCRIPTOR' : _STRINGDATAPOINT,
  '__module__' : 'data_points_pb2'
  # @@protoc_insertion_point(class_scope:com.cognite.v1.timeseries.proto.StringDatapoint)
  })
_sym_db.RegisterMessage(StringDatapoint)

StringDatapoints = _reflection.GeneratedProtocolMessageType('StringDatapoints', (_message.Message,), {
  'DESCRIPTOR' : _STRINGDATAPOINTS,
  '__module__' : 'data_points_pb2'
  # @@protoc_insertion_point(class_scope:com.cognite.v1.timeseries.proto.StringDatapoints)
  })
_sym_db.RegisterMessage(StringDatapoints)

AggregateDatapoint = _reflection.GeneratedProtocolMessageType('AggregateDatapoint', (_message.Message,), {
  'DESCRIPTOR' : _AGGREGATEDATAPOINT,
  '__module__' : 'data_points_pb2'
  # @@protoc_insertion_point(class_scope:com.cognite.v1.timeseries.proto.AggregateDatapoint)
  })
_sym_db.RegisterMessage(AggregateDatapoint)

AggregateDatapoints = _reflection.GeneratedProtocolMessageType('AggregateDatapoints', (_message.Message,), {
  'DESCRIPTOR' : _AGGREGATEDATAPOINTS,
  '__module__' : 'data_points_pb2'
  # @@protoc_insertion_point(class_scope:com.cognite.v1.timeseries.proto.AggregateDatapoints)
  })
_sym_db.RegisterMessage(AggregateDatapoints)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
